#include <string>
#include <map>
#include <boost/shared_ptr.hpp>
#ifndef _WIN32
#include <signal.h>
#include <string.h>
#include <unistd.h>
#endif


#include <log4cplus/logger.h>
#include <log4cplus/loglevel.h>
#include <log4cplus/configurator.h>
#include <log4cplus/layout.h>
#ifdef _WIN32
#include "XGetopt.h"
#include <log4cplus/nteventlogappender.h>
#else
#include <log4cplus/syslogappender.h>
#endif

bool load_config(const std::string & filename,  std::string & endpoint, std::string & api_key, std::string & feed, std::map<std::string, int> & datasets, int & update_frequency, int & min_values, std::string & msg_format);
void process_messages(const std::string & endpoint, const std::string & api_key, const std::string & feed,  std::map<std::string, int> & datasets, int update_frequency, int min_values, const std::string & msg_format);
void cleanup();
#ifdef _WIN32
void service_main(int argc, char **argv);
#endif

#ifndef _WIN32
sig_atomic_t run = 1;
#else
int run = 1;
#endif

std::string app_name("pachube_agent");
const std::string version("0.01");
std::string endpoint;
std::string api_key;
std::string feed;
std::map<std::string, int> datasets;
int update_frequency=60;
int min_values=5;

void cleanup()
{

}

#ifndef _WIN32
void sig_handler(int signum)
{
	log4cplus::Logger logger=log4cplus::Logger::getRoot();
	bool terminate = false;
	switch (signum) {
		case SIGHUP :
			LOG4CPLUS_INFO(logger, "received SIGHUP");
			break;
		case SIGTERM:
			LOG4CPLUS_INFO(logger, "received SIGTERM");
			terminate=true;
			break;
		case SIGINT:
			LOG4CPLUS_INFO(logger, "received SIGINT");
			terminate=true;
			break;
		default:
			std::ostringstream log_message;
			log_message << "unhandled signal (" << strsignal(signum) << ")";
			LOG4CPLUS_INFO(logger, log_message);
			break;
	}
	if(terminate) {
		LOG4CPLUS_INFO(logger, "closing");
		run=0;
	}
}

#endif



int main(int argc, char **argv)
{
	int opt;
	bool daemonise = false;
	std::string config_file;
	std::string logger_config;
#ifndef _WIN32
	config_file="/etc/pachube_agent/config/pachube_agent.conf";
	logger_config="/etc/pachube_agent/config/logger.conf";
#else
	config_file="C:\\Documents and Settings\\All Users\\Application Data\\envi\\pachube_agent\\pachube_agent.conf";
	logger_config="C:\\Documents and Settings\\All Users\\Application Data\\envi\\pachube_agent\\logger.conf";
#endif

	while((opt=getopt(argc, argv, "df:"))!=-1){
		switch(opt) {
			case 'd':
				daemonise = true;
				break;
			case 'f':
				config_file = optarg;
				break;
			case 'l':
				logger_config = optarg;
				break;
		}
	}
	log4cplus::PropertyConfigurator::doConfigure(logger_config);
	log4cplus::Logger logger=log4cplus::Logger::getRoot();
	std::ostringstream log_message;
	std::string msg_format="json-multi";
	log_message << "Version " << version;
	LOG4CPLUS_INFO(logger, log_message.str());
	LOG4CPLUS_INFO(logger, "starting");
	if(!load_config(config_file, endpoint, api_key, feed, datasets, update_frequency, min_values, msg_format)) {
		exit(EXIT_FAILURE);
	}
	log_message.str("");
	log_message << "endpoint: " << endpoint;
	LOG4CPLUS_INFO(logger, log_message.str());
	log_message.str("");
	log_message << "message format: " << msg_format;
	LOG4CPLUS_INFO(logger, log_message.str());
	log_message.str("");
	log_message << "feed: " << feed;
	LOG4CPLUS_INFO(logger, log_message.str());
	log_message.str("");
	log_message << "update frequency: " << update_frequency << " seconds";
	LOG4CPLUS_INFO(logger, log_message.str());
	log_message.str("");
	log_message << "minumum number of values for update: " << min_values;
	LOG4CPLUS_INFO(logger, log_message.str());
	for(std::map<std::string, int>::const_iterator itr=datasets.begin(); itr!=datasets.end(); ++itr) {
		log_message.str("");
		log_message << "sensor " << itr->first << " --> dataset " << itr->second;
		LOG4CPLUS_INFO(logger, log_message.str());
	}

#ifdef _WIN32
	if(daemonise) {
		SERVICE_TABLE_ENTRY service_table[2];
		service_table[0].lpServiceName = "rrdAgent";
		service_table[0].lpServiceProc = (LPSERVICE_MAIN_FUNCTION)service_main;
		service_table[1].lpServiceName = NULL;
		service_table[1].lpServiceProc = NULL;
		LOG4CPLUS_INFO(logger, "starting service");
		StartServiceCtrlDispatcher(service_table);
		exit(EXIT_SUCCESS);
	}
	else {
		process_messages(endpoint, api_key, feed, datasets, update_frequency, min_values, msg_format);
	}
#else
	signal(SIGHUP, sig_handler);
	signal(SIGTERM, sig_handler);
	signal(SIGINT, sig_handler);
	signal(SIGQUIT, sig_handler);
	if(daemonise) {
		LOG4CPLUS_INFO(logger, "run as daemon");
		int error=daemon(0, 0);
		if(error) {
            LOG4CPLUS_INFO(logger, "run as daemon failed");
            exit(EXIT_FAILURE);
		}
	}
	process_messages(endpoint, api_key, feed, datasets, update_frequency, min_values, msg_format);
#endif

	LOG4CPLUS_INFO(logger, "closed");
	logger.shutdown();
	exit(EXIT_SUCCESS);
}

