#include <string>
#include <boost/shared_ptr.hpp>
#include <boost/xpressive/xpressive.hpp>
#include <zmq.hpp>

#include <log4cplus/logger.h>
#include <log4cplus/loglevel.h>
#include <log4cplus/configurator.h>
#include <log4cplus/layout.h>

#include <json/value.h>
#include <json/reader.h>

#include <curl/curl.h>


#ifndef _WIN32
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
extern sig_atomic_t run;
#else
extern int run;
#endif

std::string  parse(const std::string & data);
std::string implode(const std::map<int, std::string> v,  const std::string sep);

std::map< std::string, double >  decode_message(const std::string & message_data, const std::string & msg_format);
std::map< std::string, double >  decode_message_json_multi(const std::string & message_data);
std::map< std::string, double >  decode_message_json_single(const std::string & message_data);


size_t read_callback(void *ptr, size_t size, size_t nmemb, std::string *data)
{
	int len = data->size();
	if(len) {
		memcpy(ptr, data->c_str(), len);
	}
	*data="";
	return len;
}

size_t write_callback(void *ptr, size_t size, size_t nmemb, void *user_data)
{
	log4cplus::Logger *logger = static_cast<log4cplus::Logger *>(user_data);
	int len = size * nmemb;
	char *str = new char[len + 1];
	memcpy(str, ptr, len);
	str[len] = 0;
	std::ostringstream log_message;
	log_message << "libcurl returned:" << str;
	LOG4CPLUS_DEBUG(*logger, log_message.str());
	delete [] str;
	return len;
}

int debug_callback(CURL * curl, curl_infotype info_type, char * data, size_t size, void * user_data)
{
	log4cplus::Logger *logger = static_cast<log4cplus::Logger *>(user_data);
	std::string info_type_string;
	switch(info_type) {
		case CURLINFO_DATA_IN:
			info_type_string = "DATA IN";
			break;
		case CURLINFO_DATA_OUT:
			info_type_string = "DATA OUT";
			break;
		case CURLINFO_HEADER_IN:
			info_type_string = "HEADER IN";
			break;
		case CURLINFO_HEADER_OUT:
			info_type_string = "HEADER OUT";
			break;
		case CURLINFO_TEXT:
			info_type_string = "INFO";
			break;
		default:
			info_type_string = "UNKNOWN";
			break;
	}
	char *str = new char[size+1];
	memcpy(str, data, size);
	str[size] = 0;
	std::ostringstream log_message;
	log_message << "libcurl:" << str << " {" << info_type_string << "]";
	LOG4CPLUS_DEBUG(*logger, log_message.str());
	delete [] str;
	return 0;
}

void process_messages(const std::string & endpoint, const std::string & api_key, const std::string & feed, std::map<std::string, int> & datasets, int update_frequency, int min_values, const std::string & msg_format)
{
	#ifndef _WIN32
	timespec tm_spec;
	#endif
	log4cplus::Logger logger=log4cplus::Logger::getRoot();
	int curl_result;
	zmq::context_t context(1);
	zmq::socket_t socket(context, ZMQ_SUB);
	socket.setsockopt(ZMQ_SUBSCRIBE,"",0);
	socket.connect(endpoint.c_str());
	std::ostringstream log_message;
	curl_global_init(CURL_GLOBAL_ALL);

	std::map< std::string, double >  channel_total;
	std::map< std::string, int > channel_count;
	std::map< std::string, double > channel_data;
	time_t last_update_time = 0;
	//std::ostringstream key;
	struct curl_slist *headers = NULL;
	std::ostringstream sstream;
	sstream << "X-PachubeApiKey: " << api_key;
	headers = curl_slist_append(headers, sstream.str().c_str());
	headers = curl_slist_append(headers, "Expect:");

	while(run) {
		zmq::message_t message;
		channel_data.clear();
		int ret = socket.recv(&message, ZMQ_NOBLOCK);
 		if(ret) {
			if(last_update_time == 0) {
				last_update_time = time(0);
			}
			std::string payload = (char *)message.data();
			channel_data = decode_message(payload, msg_format);
		}
		if(!ret && zmq_errno()!=EAGAIN) {
			LOG4CPLUS_DEBUG(logger, zmq_strerror(zmq_errno()));
#ifdef _WIN32
			Sleep(1000);
#else
			tm_spec.tv_sec=1;
			nanosleep(&tm_spec, NULL);
#endif
		}
		if(last_update_time==0) {
#ifdef _WIN32
			Sleep(100);
#else
			tm_spec.tv_sec=0;
			tm_spec.tv_nsec=100000000; //100ms
			nanosleep(&tm_spec, NULL);
#endif
			continue;
		}
		if(time(NULL) - last_update_time < update_frequency) {
			for(std::map<std::string, double>::const_iterator itr=channel_data.begin(); itr!=channel_data.end(); ++itr) {
				channel_total[itr->first]+=itr->second;
				channel_count[itr->first]++;
			}
#ifdef _WIN32
			Sleep(100);
#else
			tm_spec.tv_sec=0;
			tm_spec.tv_nsec=100000000; //100ms
			nanosleep(&tm_spec, NULL);
#endif
			continue;
		}
		last_update_time = time(NULL);
		std::map<int, std::string> put_data_list;
		int num_channels = 0;
		for(std::map<std::string, int>::const_iterator itr = datasets.begin(); itr!=datasets.end(); ++itr) {
			if(itr->second > num_channels)
				num_channels=itr->second;
		}
		for(int i=0; i<num_channels; i++) {
			put_data_list[i] = "";
		}

		for(std::map<std::string, double>::const_iterator itr=channel_total.begin(); itr!=channel_total.end(); ++itr) {
			if(channel_count[itr->first] >= min_values) {
				log_message.str("");
				std::string key = itr->first;
				double average = itr->second / channel_count[key];
				log_message << "key " << key << " average = " << average << " (" << channel_count[key] << " values)";
				LOG4CPLUS_DEBUG(logger, log_message.str());
				int stream_no = datasets[key];
				std::ostringstream value;
				value << average;
				put_data_list[stream_no] = value.str();
			}
			else {
				log_message.str("");
				log_message << "insufficient data ("  << channel_count[itr->first] << " values)";
				LOG4CPLUS_DEBUG(logger, log_message.str());
			}
		}
		std::string put_data = implode(put_data_list, ",");
		CURL *curl = curl_easy_init();
		curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
		curl_easy_setopt(curl, CURLOPT_URL, feed.c_str());
		curl_easy_setopt(curl, CURLOPT_VERBOSE,1L);
		curl_easy_setopt(curl, CURLOPT_UPLOAD,0L);
		curl_easy_setopt(curl, CURLOPT_PUT,1L);
		curl_easy_setopt(curl, CURLOPT_READFUNCTION, read_callback);
		curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION,write_callback);
		curl_easy_setopt(curl, CURLOPT_DEBUGFUNCTION,debug_callback);
		curl_easy_setopt(curl, CURLOPT_DEBUGDATA, &logger);
		curl_easy_setopt(curl, CURLOPT_WRITEDATA, &logger);
		long data_len = put_data.size();
		curl_easy_setopt(curl, CURLOPT_INFILESIZE, data_len);
		curl_easy_setopt(curl, CURLOPT_READDATA, &put_data);
		std::string sent_data(put_data);
		curl_result=curl_easy_perform(curl);
		if(!curl_result) {
			log_message.str("");
			log_message << "sent:"  << sent_data;
			LOG4CPLUS_DEBUG(logger, log_message.str());
		}
		else {
			log_message.str("");
			log_message << "curl error:"  << ret;
			LOG4CPLUS_ERROR(logger, log_message.str());
		}
		curl_easy_cleanup(curl);
		channel_total.clear();
		channel_count.clear();
		if(channel_data.size()) {
			for(std::map<std::string, double>::const_iterator itr=channel_data.begin(); itr!=channel_data.end(); ++itr) {
				channel_total[itr->first]+=itr->second;
				channel_count[itr->first]++;
			}
		}
	}
	curl_slist_free_all(headers);
	curl_global_cleanup();
}

std::map< std::string, double > decode_message(const std::string & message_data, const std::string & msg_format)
{
	std::map< std::string, double > result;
	if (msg_format == "json-multi") {
		return decode_message_json_multi(message_data);
	}
	if(msg_format == "json-single") {
		return decode_message_json_single(message_data);
	}
	return result;
}

std::map< std::string, double > decode_message_json_single(const std::string & message_data)
{
	std::map< std::string, double > result;
	log4cplus::Logger logger=log4cplus::Logger::getRoot();

	std::ostringstream log_message;
	std::ostringstream key;

	//LOG4CPLUS_DEBUG(logger, message_data.c_str());
	Json::Reader reader;

	std::string json = parse(message_data);
	if(json.length()>0) {
		Json::Value root;
		if(!reader.parse(json, root)) {
			log_message.str("");
			log_message << "json parse failed (" << json << ")";
			LOG4CPLUS_DEBUG(logger, log_message.str());
			return result;
		}
		std::string logger_code = root["logger"].asString();
		std::string time = root["time"].asString();
		std::string monitor_code = root["monitor"]["code"].asString();
		int channel_number = root["sensor"]["channel"].asInt();
		std::string channel_type = root["sensor"]["type"].asString();
		std::string channel_value = root["sensor"]["value"].asString();

		key.str("");
		key << logger_code << ":" << monitor_code << ":" << channel_number;

		result[key.str()]=atof(channel_value.c_str());
        //log_message.str("");
		//log_message << key.str() << "==>" << channel_value.c_str();
		//LOG4CPLUS_DEBUG(logger, log_message.str());
	}
	return result;
}

std::map< std::string, double >  decode_message_json_multi(const std::string & message_data)
{
	std::map< std::string, double > result;
	log4cplus::Logger logger=log4cplus::Logger::getRoot();
	std::ostringstream log_message;
	std::ostringstream key;

	//LOG4CPLUS_DEBUG(logger, message_data.c_str());
	Json::Reader reader;
	std::string json = parse(message_data);
	if(json.length()>0) {
		Json::Value root;
		if(!reader.parse(json, root)) {
			log_message.str("");
			log_message << "json parse failed (" << json << ")";
			LOG4CPLUS_DEBUG(logger, log_message.str());
			return result;
		}
		std::string logger_code = root["logger"].asString();
		std::string time = root["time"].asString();
		Json::Value monitors = root["monitor"];
		for(uint index1 = 0; index1!= monitors.size(); ++index1) {
			std::string monitor_code = monitors[index1]["code"].asString();
			Json::Value sensors = monitors[index1]["sensor"];
			for(uint index2=0; index2!=sensors.size(); ++index2) {
				int channel_number = sensors[index2]["channel"].asInt();
				std::string channel_type = sensors[index2]["type"].asString();
				std::string channel_value = sensors[index2]["value"].asString();

				if(channel_value!="U") {
					key.str("");
					key << logger_code << ":" << monitor_code << ":" << channel_number;
					result[key.str()]=atof(channel_value.c_str());
				}
			}
		}
	}
	return result;
}

std::string  parse(const std::string & data)
{
	std::string json;
	boost::xpressive::sregex rex = boost::xpressive::sregex::compile( "^[a-zA-Z0-9]*({.*)" );
	boost::xpressive::smatch what;
	if( boost::xpressive::regex_match( data, what, rex ) ) {
		json =  what[1].str();
	}
	return json;
}

std::string implode(const std::map<int, std::string> v,  const std::string sep)
{
	std::ostringstream stream;
	for(std::map<int, std::string>::const_iterator itr = v.begin(); itr != v.end(); ++itr) {
		stream << itr->second << sep;
	}
	std::string result = stream.str().substr(0, stream.str().size() - sep.size());
	return result;
}
